package main

import (
	"context"
	"fmt"
	"html/template"
	"log"
	"net/http"
	"os"
	"os/signal"
	"regexp"
	"strings"

	"golang.org/x/sys/unix"
)

const ListenPort = 4140

var NextURIFmt = "https://gitlab.com/idio.link/go/%s.git"

const responseTmpl = `<!doctype html>` +
	`<html>` +
	`<head>` +
	`<meta name="go-import"` +
	` content="{{.Module}} git {{.URI}}">` +
	`<meta name="go-source"` +
	` content="{{.Module}}` +
	` {{print .URI}}` +
	` {{.URI | trimSuffix ".git"}}/-/tree/master{/dir}` +
	` {{.URI | trimSuffix ".git"}}/-/blob/master{/dir}/{file}#L{line}">` +
	`</head>` +
	`<body>` +
	`go get https://{{.Module}}` +
	`</body>` +
	`</html>`

func mainLoop() int {
	srv := &http.Server{Addr: fmt.Sprintf(":%d", ListenPort)}
	cleanupChan := make(chan struct{}) // wait for clean-up
	// Catch signals
	go func() {
		sigChan := make(chan os.Signal, 1)
		defer close(sigChan)
		signal.Notify(sigChan, unix.SIGINT, unix.SIGTERM)
		sig := <-sigChan
		switch sig {
		case unix.SIGINT, unix.SIGTERM:
			err := srv.Shutdown(context.Background())
			if err != nil {
				log.Printf("shutdown: %w", err)
			}
			close(cleanupChan)
		}
	}()
	// Prep template
	tmpl, err :=
		template.
			New("resp").
			Funcs(template.FuncMap{
				"trimSuffix": func(s, in string) string {
					return strings.TrimSuffix(in, s)
				},
			}).
			Parse(responseTmpl)
	if err != nil {
		log.Printf("BUG: parse template: %w", err)
		return 1
	}
	re := regexp.MustCompile(`/v[0-9]+$`)
	// Set handler
	http.HandleFunc(
		"/go/",
		func(w http.ResponseWriter, req *http.Request) {
			q := req.URL.Query()
			if q.Get("go-get") != "1" {
				code := 404
				http.Error(w, http.StatusText(code), code)
				return
			}
			modPath :=
				re.ReplaceAllString(req.Host+req.URL.Path, "")
			modName :=
				strings.TrimPrefix(modPath, req.Host+"/go/")
			err := tmpl.Execute(w, struct {
				Module string
				URI    string
			}{modPath, fmt.Sprintf(NextURIFmt, modName)})

			if err != nil {
				code := 500
				http.Error(w, http.StatusText(code), code)
			}
			return
		},
	)
	err = srv.ListenAndServe()
	if err != http.ErrServerClosed {
		log.Printf("FATAL: %w", err)
	}
	<-cleanupChan
	return 1
}

func main() {
	os.Exit(mainLoop())
}
